function [distdata, testinfo] = knnonectrl(controls, dataset, numneigh, distsize, pvalue, metric, plotdist, distctrls)
% knnonectrl is used in order to decide if a given data set is separate
% from the controls in space.  The function first randomly extracts a set
% of controls of size "distsize" for which to measure against the rest of
% the controls, or benchmark controls.  The average distances of the
% dataset and the distribution controls to the "numneigh" nearest benchmark
% controls is calculated.  KS-test is performed on the distribution
% controls distances and the dataset distances in order to decide if they
% are from the same parent distribution.  testinfo.h = 1 implies that the
% distributions do not come from the same parent distribution and that the
% median of the dataset distances is greater than the median of the control
% distances - suggesting the dataset is separate from the control.
% control object to its k(=numneigh) nearest control objects and then
% measures the average distance from each object in the dataset to its k
% nearest control objects.  Then the Pvals for the distances for each
% object in the dataset are computed using the control distances 
% distribution.
% SYNTAX: 
% [distdata, testinfo] = knnonectrl(controls, dataset, numneigh, distsize, pvalue, metric, plotdist, distctrls)
% INPUTS:   controls = the control set
%           dataset = the dataset for which the Pvals are measured
%           numneigh = the number of nearest neighbours to use
%           distsize = size of the distribution to use (control case)
%           pvalue = p-value to use in the KS-test, default = 0.05;
%           metric = the metric to use, default = 'euclidean', possible
%                    values: 'euclidean', 'seuclidean', 'mahalanobis',
%                    'cityblock', 'minkowski', 'cosine', 'correlation',
%                    'spearman', 'hamming', 'jaccard', 'chebychev'; for
%                    more information see Matlab help for pdist
%           plotdist = shows the distribution of distances of the controls
%                      and the distribution of distances of the dataset
%                      possible values: 'yes' or 'no', default = 'no'
%           distctrls = distribution controls, this is only used if the
%                       user sets distsize = 0, then the user can specify 
%                       the distribution controls, leaving "controls" as the
%                       benchmark controls to measure against.
% OUTPUTS:  distdata.ctrldistances = distances to benchmark of the dataset
%           distdata.testdistances = distances to benchmark of the controls
%           distdata.testpvals = pvalue based on the created control
%                                distribution computer for each dataset 
%                                value
%           testinfo.normhistogram = histogram of distance data
%           testinfo.ctrlmedian = median of the control distances
%           testinfo.datamedian = median of the dataset distances
%           testinfo.pvalue = KS test p-value
%           testinfo.ksstat = KS test statistic;
%           testinfo.h = hypothesis test; 1 rejects the null hypothesis the
%                        the distance distributions are from the same
%                        parent distribution and suggests the distributions
%                        are different.  0 accepts the null hypothesis that
%                        the distributions are from the same parent
%                        distribution under the condition that the median
%                        of the control distances is less than the median
%                        of the dataset distances.  A warning will appear
%                        in testinfo if the distributions are not from the
%                        same parent distribution but the control distance
%                        median is greater than the dataset distances.

% created by: Sean Cianflone, trained monkey
% ideas: Glauber Brito, Ph.D., monkey trainer
% Feb 6, 2010
% Clapton is God. 
global plotstr
%% defaults
if(nargin < 7)
    plotdist = 'no';
    if(nargin < 6)
        metric = 'euclidean';
        if(nargin<5)
            pvalue = 0.05;
        end
    end
end
%% error messages
if(nargin==8)
    if(distsize > 0)
        error('To use the specified the distribution controls, one must set distsize = 0')
    end
end
if(numel(controls(:,1))<distsize)
    error('There are not enough controls for that distribution size.')
end
if(numel(controls(:,1)) - distsize < numneigh)
    error('There are not enough controls for that distribution size and that many neighbours.')
end
%% function
%% separating distribution controls from the controls for measuring against
r = randperm(numel(controls(:,1)));
if(distsize == 0)
    distcontrols = distctrls;%the controls used to create the distribution, this additions allows you to pick the distribution controls, leaving the rest of the controls as benchmarks
else
    distcontrols = controls(r(1:distsize), :);%the controls used to create the distribution
end
meascontrols = controls(r(distsize+1:end), :); %the benchmarks to measure everything against
numctrl = numel(meascontrols(:,1));
clear controls
%% calculating pairwise distances between distcontrols and meascontrols
alldata = vertcat(meascontrols, distcontrols);%concatenating all the data for pdist
ctrlgps = vertcat(zeros(numel(meascontrols(:,1)), 1), ones(numel(distcontrols(:,1)), 1));%setting up groups, 0 = meascontrols, 1 = distcontrols
%adjusting the ctrlds data
ctrlds = pdist(alldata, metric); % calculating all pairwise distances, this is overkill, but probably faster than setting up each individual pdist
ctrlds = squareform(ctrlds); % changes the setup to be a matrix.... for ease of use
ctrlds = ctrlds(ctrlgps == 1, 1:numctrl); % control distances from the meascontrol benchmark
ctrlds = sort(ctrlds,2); % sorts the distances smallest to largest
ctrlds = ctrlds(:, 1:numneigh); % finds the nearest neighbour distances
ctrlds = mean(ctrlds, 2);% average distance to nearest "numneigh" control objects
%% calculating pairwise distances between dataset and meascontrols
alldata = vertcat(meascontrols, dataset);%concatenating all the data for pdist
datagps = vertcat(zeros(numel(meascontrols(:,1)), 1), ones(numel(dataset(:,1)), 1));%setting up groups, 0 = meascontrols, 1 = dataset
%adjusting the datads data
datads = pdist(alldata, metric); % calculating all pairwise distances, this is overkill, but probably faster than setting up each individual pdist
datads = squareform(datads); % changes the setup to be a matrix.... for ease of use
datads = datads(datagps == 1, 1:numctrl); % dataset distances from the meascontrol benchmark
datads = sort(datads,2); % sorts the distances smallest to largest
datads = datads(:, 1:numneigh); % finds the nearest neighbour distances
datads = mean(datads, 2);% average distance to nearest "numneigh" control objects
%% creating the distributions
distances = sortrows(vertcat(datads, ctrlds)); %all distances, sorted smallest to largest
dnums = ceil(0.95*numel(distances));
maxy = max(distances(1:dnums, :));%max possible distance (all data), max of top 95% of data, display purposes only!!!!!!!
miny = min(distances);%min possible distance (all data)
bins = (miny:(maxy-miny)/19:maxy)';% bins for the histogram, 20 points
testh = histc(datads, bins)/numel(datads);%dataset distances histogram (normalized to size of set)
ctrlh = histc(ctrlds, bins)/numel(ctrlds);%control distances histogram (normalized to size of set)
%% computing the probabilities based on the distribution created using ctrlds
Pvals = [];
zzz = size(distcontrols);
for i = 1:numel(datads)
    f = find(ctrlds>datads(i));    
    Pvals = vertcat(Pvals, numel(f)/zzz(1));
end
%% performing KS-test on the distributions of distcontrol distances versus dataset distances
[h,p,ks2stat] = kstest2(ctrlds, datads);
%% creating the distribution if necessary
if (strcmpi(plotdist, 'yes') == 1)
    %plotting the histogram
    figure
    plot(bins, ctrlh, 'color', [0 0 1])
    hold on
    plot(bins, testh, 'color', [1 0 0])
    hold on
    m = 0.65*max(vertcat(ctrlh, testh));
    texty = sprintf('KS p-value = \n %d', p);
    text(bins(14), m, texty)%plots the p-value on the graph
    hold off
    titlestr = sprintf('Distribution of Average Distance to the Nearest Neighbours (%s)',plotstr);
    title(titlestr)
    xlabel('Average Distance')
    ylabel('Relative Frequency')
    legend('Control Distances', 'Data Distances')
end

%% distdata outputs
distdata.ctrldistances = ctrlds; % distances to benchmark of the dataset
distdata.testdistances = datads; % distances to benchmark of the dataset
distdata.testpvals = Pvals; % pvalue based on the created control distribution computer for each dataset value
%% testinfo outputs
testinfo.normhistogram = vertcat({'Bins', 'Control Distances', 'Test Distances'}, num2cell(horzcat(bins, ctrlh, testh)));%exported histogram data (normalized)
testinfo.ctrlmedian = median(ctrlds);
testinfo.datamedian = median(datads);
testinfo.pvalue = p;% pvalue from KStest
testinfo.ksstat = ks2stat;
testinfo.h = 0; % ctrls and dataset are in the same place, until the next line
if(p<pvalue) && (testinfo.ctrlmedian<testinfo.datamedian) % if p<pvalue, then reject hypoth dist's come from same parent; ctrlmed<datamed means data set is farther from controls - ie. different
    testinfo.h = 1;
end
if(p<pvalue) && (testinfo.ctrlmedian>testinfo.datamedian) % here if datamed < ctrlmed then we have a tight/spreadout cluster inside the normals - ie. different distribution shape, but inside the control cluster
    testinfo.warning = 'This cluster may be inside the control cluster but the distribution is not the same.';
end
end